'use strict';

// utm 参数白名单，如果小程序页面参数中的这些参数不添加 p_query 前缀而是添加 p_ 前缀
// 如 utm_campaign => p_utm_campaign
const utmWhiteList = ['utm_campaign', 'utm_source', 'utm_medium', 'utm_term', 'utm_content'];

function assignQueryParams(source, query) {
  Object.keys(query).forEach(key => {
    if (utmWhiteList.includes(key)) {
      source[key] = query[key];
    } else {
      source[`query_${key}`] = query[key];
    }
  });
}

const __debug = (...args) => console.log('[hxt] ', ...args);

const HXT_SDK_VERSION = '0.1.8';

const hxtConfig = require('./app.json').hxt;

function __setStorage(key, value) {
  return my.setStorageSync({
    key,
    data: value,
  });
}

if (!hxtConfig) {
  __debug('app.json 缺少必要的 hxt 配置');

  module.exports = {
    App,
    Page
  };
}
if (hxtConfig) {
  __debug('hxt config detected.', hxtConfig);

  const defaultConfig = {
    module: {
      basic: {
        apiUrl: 'https://t.hypers.com.cn/hxt'
      }
    }
  };

  let _initialized = false;

  const hxt = {
    $identity: {},
    $config: {
      ...defaultConfig,
      ...hxtConfig
    },
    $client: {},
    $pages: {},
    $page: ''
  };

  hxt.App = options => App({
    hxt,
    ...options,
    onLaunch(...args) {
      hxt.init(args[0]);
      return options.onLaunch && options.onLaunch.call(this, ...args)
    },
    onShareAppMessage: options.onShareAppMessage && function (...args) {
      const shareObj = options.onShareAppMessage.call(this, ...args);
      shareObj && hxt.trackAction('share', { share_title: shareObj.title, share_path: shareObj.path });
      return shareObj
    }
  });

  hxt.Page = ({ name: $pageName, ...options }) => {
    if (!$pageName) {
      __debug('hxt.Page 缺少必要选项 `name`');
    } else if (hxt.$pages[$pageName]) {
      __debug(`hxt.Page.name 重复: ${$pageName}`);
    }

    Page({
      ...options,
      onLoad(query, ...args) {
        hxt.trackPageStart($pageName, query);
        return options.onLoad && options.onLoad.call(this, query, ...args)
      },
      onShow(...args) {
        hxt.trackPageStart($pageName);
        return options.onShow && options.onShow.call(this, ...args)
      },
      onHide(...args) {
        hxt.trackPageEnd($pageName);
        return options.onHide && options.onHide.call(this, ...args)
      },
      onUnload(...args) {
        hxt.trackPageEnd($pageName);
        return options.onUnload && options.onUnload.call(this, ...args)
      },
      onShareAppMessage: options.onShareAppMessage && function (...args) {
        const shareObj = options.onShareAppMessage.call(this, ...args);
        shareObj && hxt.trackAction('share', { share_title: shareObj.title, share_path: shareObj.path });
        return shareObj
      }
    });
  };

  /**
   * @param {query, path, scene, referrerInfo} options
   */
  hxt.init = (options) => {
    const launchParams = {
      path: options.path,
      scene: options.scene
    };
    if (options.query) {
      assignQueryParams(launchParams, options.query);
    }
    if (options.referrerInfo) {
      launchParams.appid = options.referrerInfo.appId;
      launchParams.ref_info = JSON.stringify(options.referrerInfo);
    }
    __report({ type: 'client_data', ...__prependP(launchParams) });

    Promise.all([
      __initIdentity(),
      __initDeviceId(),
      __initSystem(),
      __initNetwork(),
      __initLocation()
    ]).then(() => {
      __doneInitialize();
    });
  };

  hxt.trackPageStart = (pageName, query = {}) => {
    if (hxt.$page !== pageName) {
      __debug('trackPageStart', pageName, query);
      hxt.$page = pageName;
      const $page = (hxt.$pages[pageName] = hxt.$pages[pageName] || {});
      $page.entryTime = Date.now();
      $page.query = {};
      assignQueryParams($page.query, query);
    }
  };

  hxt.trackPageEnd = pageName => {
    const $page = (hxt.$pages[pageName] = hxt.$pages[pageName] || {});
    hxt.$page = '';
    const $exitTime = Date.now();
    const stayLength = $exitTime - $page.entryTime;
    __debug('trackPageEnd', pageName, `stayed for ${stayLength}ms`);
    __report({ type: 'activity', activity: pageName, start_ts: $page.entryTime, end_ts: $exitTime, duration: stayLength, ...__prependP($page.query) });
  };

  hxt.trackAction = (name, props) => {
    __report({ type: 'act', activity: hxt.$page, act_name: name, ...__prependP(props) });
  };

  hxt.identify = (identity) => {
    hxt.$identity = { ...hxt.$identity, ...identity };
    __setStorage('hxt.$identity', hxt.$identity);
    if (hxt.$identity.user_id) {
      __start();
    }
  };

  function __initIdentity() {
    const storageKey = 'hxt.$identity';

    return new Promise(resolve => {
      my.getStorage({
        key: storageKey, // 缓存数据的key
        success: ({ data }) => {
          __debug('identity', data);
          hxt.$identity = data || {};
        },
        complete: resolve
      });
    })
  }

  function __initDeviceId() {

    const storageKey = 'device_id';

    return new Promise(resolve => {
      my.getStorage({
        key: storageKey, // 缓存数据的key
        success: ({ data }) => {
          let device_id = data;

          if (!!device_id) {
            __debug('device_id', device_id);
            hxt.$client = { ...hxt.$client, device_id };
            resolve();
          } else {
            device_id = uuidv4();
            my.setStorage({
              key: storageKey, // 缓存数据的key
              data: device_id, // 要缓存的数据
              success: () => {
                __debug('device_id', device_id);
                hxt.$client = { ...hxt.$client, device_id };
              },
              complete: resolve
            });
          }
        },
      });
    })
  }

  function __initSystem() {
    return new Promise((resolve) => {
      my.getSystemInfo({
        success($system) {
          __debug('system', $system);
          hxt.$client = { ...hxt.$client, ...$system };
        },
        complete: resolve
      });
    })
  }

  function __initNetwork() {
    return new Promise(resolve => {
      my.getNetworkType({
        success(res) {
          __debug('network', res);
          hxt.$client = { ...hxt.$client, ...res };
        },
        complete: resolve
      });
    })
  }

  function __initLocation() {
    if (!hxt.$config.locationAccessible) {
      return Promise.resolve()
    }
    return new Promise(resolve => {
      my.getLocation({
        success(res) {
          __debug('location', res);
          hxt.$client = {
            ...hxt.$client,
            lon: res.longitude,
            lat: res.latitude
          };
        },
        complete: resolve
      });
    })
  }

  function __doneInitialize() {
    _initialized = true;
    __debug('hxt initialized.');
    __start();
  }

  const _queuedReports = [];

  let _started = false;

  // 尝试开始处理队列的请求
  function __start() {
    if (_started) return;
    if (!_initialized) return;
    if (hxtConfig.requireUserId && !hxt.$identity.user_id) return;

    _started = true;
    __consumeQueuedReports();
  }

  function __consumeQueuedReports() {
    _queuedReports.length && __debug('digesting queued reports.');
    while (_queuedReports.length) {
      __doReport(_queuedReports.shift());
    }
  }
  function __report(params) {
    if (!_started) {
      __debug('queued report', params);
      _queuedReports.push(params);
      return
    }

    __doReport(params);
  }

  function __doReport(params) {
    const reportData = __makeReportData(params);

    __debug('report', reportData.type, reportData);
    console.info(`[hxt] 数据上报：${{
      client_data: '初始化',
      activity: '页面活动',
      act: '自定义事件'
    }[reportData.type]}`, reportData);
    try {
      my.request({
        url: `${hxt.$config.module.basic.apiUrl}?_ua=${encodeURIComponent(hxt.$config.module.basic.appKey)}&_t=i&_z=m`, // 目标服务器url
        headers: {
          'content-type': 'application/x-www-form-urlencoded'
        },
        method: 'POST',
        data: reportData,
        dataType: 'arraybuffer',
        success: (res) => {
          __debug('report success', res);
        },
      });
    } catch (e) {
      my.httpRequest({
        url: `${hxt.$config.module.basic.apiUrl}?_ua=${encodeURIComponent(hxt.$config.module.basic.appKey)}&_t=i&_z=m`, // 目标服务器url
        method: 'POST',
        data: reportData,
        success: (res) => {
          __debug('report success', res);
        },
      });
    }
  }

  function __makeReportData({ type, act_name, activity, ...params } = {}) {

    const { $config, $client, $identity: { user_id: alipay_id, cust_id, ...userInfo } } = hxt;

    // os
    let os = 3;
    const isIOS = /ios|iphone/i.test($client.platform);
    if (isIOS) {
      os = 1;
    } else if (/android/i.test($client.platform)) {
      os = 0;
    }

    const [width, height] = $client.screenWidth ?
      [$client.screenWidth, $client.screenHeight] :
      [$client.windowWidth, $client.windowHeight];

    const sr = isIOS ?
      `${width * $client.pixelRatio}x${height * $client.pixelRatio}` :
      `${width}x${height}`;

    // device_name, model, manufacturer
    const manufacturer = $client.brand;
    const model = $client.model;
    const device_name = (model && model.includes(manufacturer)) ? model : `${manufacturer} ${model}`;

    return {
      _ua: $config.module.basic.appKey,
      device_id: hxtConfig.requireUserId ? alipay_id : $client.device_id,
      type,
      act_name,
      activity,
      _z: 'm',
      ts: Date.now(),
      app_name: $config.appName,
      app_code: $config.appCode,
      app_version: $config.appVersion,
      os,
      os_version: $client.system,
      channel_id: '支付宝小程序',
      device_name,
      model: isIOS ? '' : model,
      manufacturer,
      network: $client.networkType.toLowerCase(),
      sr,
      lang: $client.language,
      p_alipaysv: $client.version,
      xcxsv: my.SDKVersion,
      alipay_id,
      cust_id,
      ...__prependP(userInfo),
      lon: $client.lon,
      lat: $client.lat,
      sv: HXT_SDK_VERSION,
      ...params
    }
  }

  function __prependP(params = {}) {
    return Object.keys(params).reduce((acc, key) => ({
      ...acc,
      [`p_${key}`]: params[key]
    }), {})
  }

  function uuidv4() {
    return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function (c) {
      var r = Math.random() * 16 | 0, v = c == 'x' ? r : (r & 0x3 | 0x8);
      return v.toString(16);
    });
  }

  module.exports = hxt;
}
